module Main(main) where

import qualified Board(load)
import Concurrent
import Monad(unless)
import Network 
import Posix(installHandler,sigPIPE,Handler(Ignore))
import System(getArgs,exitFailure)
import GetOpt
import IO(stderr,hPutStrLn)
import Utils

import qualified Server


--------------------------------------------------------------------------------
-- command line options
--------------------------------------------------------------------------------
data GameType = Continuous | OneGame Int | ManyGames Int

data Option = Opt
  { portNum   :: PortID 
  , gameType  :: GameType
  , mapFile   :: Maybe String
  , packFile  :: Maybe String
  , fuel      :: Integer
  , capacity  :: Integer
  } 

default_options = Opt
  { portNum   = PortNumber 20005
  , gameType  = Continuous
  , mapFile   = Nothing
  , packFile  = Nothing
  , fuel      = 5000
  , capacity  = 50
  }

options = 
  [ Option ['p'] ["port"] (ReqArg (num port) "NUM")   
    "port to listen on"

  , Option ['n'] ["new"] (ReqArg (num manyGames) "NUM")
    "new game every NUM players"

  , Option ['o'] ["one"] (ReqArg (num oneGame) "NUM")
    "one game with NUM players"

  , Option ['m'] ["map"] (ReqArg mapF "FILE")   
    "the map to use for the game"

  , Option ['k'] ["packs"] (ReqArg packF "FILE")
    "locations of packages"

  , Option ['f'] ["fuel"] (ReqArg (num setFuel) "NUM") 
    "fuel for each player"

  , Option ['c'] ["capacity"] (ReqArg (num setCapa) "NUM") 
    "capacity for each player"
  ]
  where
  num f x opts = 
    case parse x of
      Nothing -> opts
      Just y -> f y opts 

  port x opts = opts { portNum = PortNumber (fromIntegral x) }

  manyGames x opts = opts { gameType = ManyGames x }
  oneGame x opts = opts { gameType = OneGame x }
  
  setFuel x opts  = opts { fuel = x }
  setCapa x opts  = opts { capacity = x }
  packF x opts    = opts { packFile = Just x }
  mapF x opts     = opts { mapFile = Just x }


  

--------------------------------------------------------------------------------
-- the driver
--------------------------------------------------------------------------------
main = withSocketsDo $
  do installHandler sigPIPE Ignore Nothing
     args <- getArgs
     let (optFs,_,errs) = getOpt Permute options args
	 opts = foldr ($) default_options optFs

     unless (null errs) showUsage
     let invalidMapFile  = putStrLn "Invalid map file"  >> exitFailure
	 invalidPackFile = putStrLn "Invalid pack file" >> exitFailure
  
     (mapF,packF) <- maybe showUsage return $ 
      do mapF <- mapFile opts
         packF <- packFile opts
         return (mapF,packF)

     board <- maybe invalidMapFile return =<< Board.load mapF
     packs <- maybe invalidPackFile return . parse =<< readFile packF

     chan <- newChan
     Server.start chan (portNum opts) board (fuel opts) (capacity opts)
     case gameType opts of
       Continuous  -> Server.contGame chan packs board
       ManyGames n -> Server.singleGames n chan packs board (portNum opts) 
       OneGame n   -> Server.oneGame n chan packs board
     

showUsage = do
  hPutStrLn stderr $ usageInfo "Required: map and packs.\n usage:" options
  exitFailure



