module Server(start,oneGame,singleGames,contGame) where

import Concurrent
import System(getEnv)
import Time
import IO
import Maybe(catMaybes)
import Monad(unless,when)
import Network hiding (accept)
import SocketPrim(accept,socketToHandle)
import Prelude hiding (log)
import Utils
import Ix
import Array

import qualified Board
import qualified Game
import qualified Observer
import qualified Pack
import qualified Robo


--------------------------------------------------------------------------------
-- the server
--------------------------------------------------------------------------------
start chan port board fuel capacity = forkIO $
  do sock <- listenOn port
     mapM_ (getClient sock) [1..]
     where
     close h = hClose h `catch` \e -> return ()

     getClient sock n = do
       h <- flip socketToHandle ReadWriteMode . fst =<< accept sock
       forkIO $ mkPlayer h n `catch` const (close h) 
     
     mkPlayer h uid = 
      do ln <- hGetLine h 
         case parse ln of
          Just (Game.Login x y) 
            | inRange (bounds board) (x,y) -> 
                do r <- Robo.create uid (x,y) fuel capacity h 
                   writeChan chan (Game.Robo r)
            | otherwise -> close h 
          Just Game.Player -> 
                do r <- Robo.create uid (1,1) fuel capacity h
                   writeChan chan (Game.Robo r)
          Just Game.Observer -> 
                do o <- Observer.create h
                   writeChan chan (Game.Obs o) 
          _ -> close h
  



--------------------------------------------------------------------------------
-- games that end
--------------------------------------------------------------------------------

-- initialization
newGame n chan packs board h = 
  do players <- sequence $ replicate n $ waitForPlayer (hPutStrLn h) chan
     g <- Game.create packs board players h
     mapM_ (initPlayer g packs) players
     return g


waitForPlayer log chan = 
  do p <- readChan chan 
     case p of 
       Game.Robo r -> 
         comment log ("a player joined. uid = " ++ show (Robo.uid r))
       Game.Obs o -> comment log "an observer joined."
     return p



-- start many games in a single server 
singleGames n chan packs board port = 
  mapM_ (oneGameOfMany n chan packs board port) [1..]

oneGameOfMany n chan packs board (PortNumber port) gameId = 
  do comment putStr $ 
        "ready for game " ++ show gameId ++ " (" ++ show n ++ " players)"
     time <- calendarTimeToString # (toCalendarTime =<< getClockTime)
     let logFileName = spacesTo_ $ "game_" ++ time ++ "_" ++ show port 
                                           ++ "_" ++ show gameId ++ ".log"
     logFile <- openFile logFileName WriteMode
     logFile `hSetBuffering` LineBuffering
     forkIO . playSingleGame chan =<< newGame n chan packs board logFile
oneGameOfMany _ _ _ _ _ _ = error "Unsupported kind of port"
    
 
-- start just a single finishing game - no forking, log to stdout
oneGame n chan packs board = 
  do comment putStrLn $ "ready for game (" ++ show n ++ " players)"
     stdout `hSetBuffering` LineBuffering
     playSingleGame chan =<< newGame n chan packs board stdout 

playSingleGame chan g = 
  do Game.oneTurn g `untilM` Game.noRobots g
     Game.over g


--------------------------------------------------------------------------------
-- a continuous game
--------------------------------------------------------------------------------

contGame chan packs board = 
  do comment putStrLn "ready for a continuous game"
     hSetBuffering stdout LineBuffering
     g <- Game.create packs board [] stdout
     playContGame chan g packs 
     

playContGame chan g packs
  = do whenM (Game.noRobots g) (newPlayers True 1)
       newPlayers False maxConns
       Game.oneTurn g
       whenM (Game.noPackages g) (g `Game.addPackages` packs)
       playContGame chan g packs
  where
  -- max players that can join each turn
  maxConns = 5

  -- check for new players
  newPlayers wait n = do
    newRs <- takeC chan wait n
    (g `upd` Game.players) (newRs ++)
    mapM_ (initPlayer g packs) newRs
    let num = length newRs
    when (num /= 0) $ 
      do comment putStrLn (show num ++  " new players joined the game." )
         addMorePacks

  addMorePacks = 
    do mx <- maximum . map Pack.uid # (g `get` Game.packages)
       let incID p = p { Pack.uid = Pack.uid p + mx + 1 }
           addGr gr@(loc,newPs) = 
              do ps <- g `Game.getPacksAt` loc
                 return $ if null ps then Just (loc, map incID newPs)
                                     else Nothing

       newpacks <- catMaybes # mapM addGr packs
       g `Game.addPackages` newpacks



--------------------------------------------------------------------------------
-- init players when they first connect
--------------------------------------------------------------------------------    
initPlayer g _ (Game.Robo r) =
  do  r `Robo.sendStrLn` Board.toString (Game.board g)
      m <- r `get` Robo.money 
      r `Robo.sendStrLn` unwords [show (Robo.uid r),
                                  show (Robo.capacity r),
                                  show m]
      infos <- mapM getInfo =<< Game.getRobots g
      r `Robo.sendStrLn` Game.mkRsp infos
      
      hFlush (Robo.talkTo r)
  `catch` const (Robo.kill "<connection was closed>" r)

  where
  getInfo s = do pos <- s `get` Robo.pos
                 return (Robo.uid s, [Robo.Spawned pos])

initPlayer g packs (Game.Obs o) = 
  do o `Observer.sendStrLn` Board.toString (Game.board g)
     o `Observer.sendStrLn` show (Game.packDests packs)
  




